'''
--------------------------------------------------------------
IRIS: (GPU-accelerated) IR spectrum modeling

Evaluation of optical depths and intensities is jit-enabled and
vectorized. 
--------------------------------------------------------------
Developed by Carlos E. Muñoz-Romero (2023)
'''
import jax
jax.config.update("jax_enable_x64", True)

import astropy.units as u
import astropy.constants as const
import jax.numpy as jnp
import numpy as np
from functools import partial

# -------------------- Global constants ----------------------
# for more efficient unit handling
# au to cm
aucm = const.au.to(u.cm)
# pc to cm
pccm = const.pc.to(u.cm)
# squared arcsec to sr
arcsec2_to_sr = ((u.arcsec**2).to(u.sr))
# c in micron/sec
c_musec = (const.c.to(u.micron/u.s)).value
# c in cm/sec
c_cmsec = (const.c.cgs).value
# km/s to micron/s
kms_to_mus = ((1*u.km/u.s).to(u.micron/u.s)).value
# plack const. in cgs
h_cgs = (const.h.cgs).value
# boltzmann const. in cgs
k_B_cgs = (const.k_B.cgs).value    
# c/micron
micron_to_hz = (const.c/(1*u.micron)).to(u.Hz)
# mass of the proton
mp_cgs = const.m_p.cgs.value
# ----------------------------------------------------------

# --------------- Main modeling code -----------------------

def J_profile(wavelength, t_ex): 
    """
    J_profile: full (blackbody) Plack function

    :wavelength: wavelength in micron
    :t_ex: temperature in K
    
    :return: blackbody flux density in erg cm-2 s-1 Hz-1 sr-1
    """ 
    nu = c_cmsec/(wavelength*1e-4)
    return (2*h_cgs*nu**3)/(c_cmsec**2) * (jnp.exp(h_cgs*nu/(k_B_cgs*t_ex)) - 1)**(-1)
    
J_profile = jax.jit(J_profile)

def dv_to_dlam(dv, lam_rest):
    """
    dv_to_dlam: convert velocity element in km/s to wavelength element in micron,
                given a reference wavelength
    :dv: elocity element in km/s
    :lam_rest: reference wavelength in micron
    
    returns: dlam in micron
    """
    nu_rest = c_musec/lam_rest
    lam =  -c_musec / (((dv*kms_to_mus*nu_rest)/c_musec) - nu_rest)
    return lam - lam_rest
dv_to_dlam = jax.jit(dv_to_dlam)

def evaluate_line_tau(fine_wgrid, sigma_lam, tau_cen, line_w):
    """
    evaluate_line_tau: get line optical depth profile 
    
    :fine_wgrid: fine wavelength grid to evaluate tau 
    :sigma_lam: intrinsic line (gaussian) width in micron
    :tau_cen: optical depth at line center
    :line_w: wavelenbgth at line center in micron
    
    returns: tau (unitless)
    """
    # Gaussian profile
    return tau_cen * jnp.exp(-((fine_wgrid-line_w)**2)/(2*sigma_lam**2) )
evaluate_line_tau = jax.jit(evaluate_line_tau)

def compute_tau_grid(line_catalog, fine_wgrid, dv, t_ex, n_mol):
    """
    compute_tau_grid: get optical depth profile for all lines in catalog
    
    :line_catalog: HITRAN catalog 
    :fine_wgrid: fine wavelength grid to evaluate tau 
    :dv: intrinsic line FWHM in km/s
    :t_ex: excitation temperature in K
    :n_mol: column density in cm^-2
    
    returns: tau (unitless)
    """
    # partition function
    q_sum = jnp.interp(t_ex, line_catalog['Qt'], line_catalog['Qv'])
    # population levels
    x_low = line_catalog['glow'] * jnp.exp(-line_catalog['elow'] / t_ex) / q_sum
    x_up = line_catalog['gups'] * jnp.exp(-line_catalog['eups'] / t_ex) / q_sum
    # optical depth
    tau_amp = (jnp.log(2) / jnp.pi)**0.5 * (line_catalog['aijs'] * n_mol * (line_catalog['ws']*1e-4)**3) / (4*jnp.pi*dv*(1e5))
    tau_lvl = x_low * line_catalog['gups']/line_catalog['glow'] - x_up
    tau_cen = tau_amp * tau_lvl
    # intrinsic width in micron
    sigma_lam = dv_to_dlam(dv/2.355, fine_wgrid)
    # map over all lines
    func = partial(evaluate_line_tau, fine_wgrid, sigma_lam)
    tau_grid = jax.vmap(func)(tau_cen, line_catalog['ws'])
    
    # sum opacities
    return  jnp.sum(tau_grid, axis=0)
compute_tau_grid = jax.jit(compute_tau_grid)

def compute_total_tau(catalog, fine_wgrid, dv, t_ex, n_mol):
    """
    compute_tau_grid: get optical depth profile for all molecules
    
    :catalog: HITRAN catalogs
    :fine_wgrid: fine wavelength grid to evaluate tau 
    :dv: intrinsic line FWHM in km/s
    :t_ex: excitation temperature in K
    :n_mol: column density in cm^-2
    
    returns: tau (unitless)
    """
    keys = list(catalog.keys())
    total_tau = []
    # loop over each species
    for i in range(len(catalog)):
        # map over disk grid
        func = partial(compute_tau_grid, catalog[keys[i]], fine_wgrid)
        total_tau.append(jax.vmap(func)(dv[i], t_ex[i], n_mol[i]))
    return jnp.array(total_tau)

compute_total_tau = jax.jit(compute_total_tau)

def compute_fdens(distance, fine_wgrid, total_tau, A_au, t_ex):
    """
    compute_fdens: get flux density for one species
    
    :distance: distance to source in pc
    :fine_wgrid: fine wavelength grid to evaluate tau 
    :total_tau: tau as function of wavelength
    :A_au: emitting area in au^2
    :t_ex: excitation temperature in K
    
    returns: flux density (erg cm-2 s-1 Hz-1)
    """
    # radius assuming circular area
    eq_radius = (A_au/jnp.pi)**0.5
    # line intensities accounting for saturation at line center
    return jnp.pi * (eq_radius/distance)**2 * ((aucm / pccm) ** 2) * J_profile(fine_wgrid, t_ex) * (1 - jnp.exp(-total_tau))
compute_fdens = jax.jit(compute_fdens)

def compute_total_fdens(catalog, distance, A_au, t_ex, n_mol, dv, fine_wgrid):
    """
    compute_total_fdens: get total flux density in Jy
    
    :catalog: HITRAN catalog (dict)
    :distance: distance to source in pc
    :A_au: emitting area in au^2
    :t_ex: excitation temperature in K
    :n_mol: column density in cm^-2
    :dv: intrinsic line FWHM in km/s
    :fine_wgrid: fine wavelength grid to evaluate tau 
   
    returns: flux density (Jy)
    """
    fdens = fine_wgrid*0.0
    # evaluate total optical depth profile
    total_tau = compute_total_tau(catalog, fine_wgrid, dv, t_ex, n_mol)
    # loop over species
    for i in range(len(catalog)):
        # map over disk grid
        func = partial(compute_fdens, distance, fine_wgrid)
        fdens_i = jax.vmap(func)(total_tau[i], A_au[i], t_ex[i])
        fdens += jnp.sum(fdens_i, axis=0) 
    return fdens * 1e23
compute_total_fdens = jax.jit(compute_total_fdens)



'''
--------------------------------------------------------------
IRIS: (GPU-accelerated) IR spectrum modeling
--------------------------------------------------------------
Developed by Carlos E. Muñoz-Romero (2023)
'''
import os
import sys

sys.path.append('./')
import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
import spectrum as sp
import numpy as np
from moldata import setup_catalog
from jax.scipy.signal import fftconvolve

class slab:
    """
        slab: generate an object to model spectra

        :wlow: shortest wavelength in micron
        :whigh:  longest wavelength in micron
        :path_to_hitran: path to folder with HITRAN molecular data
        
    """
    def __init__(self, molecules, wlow, whigh):
        
        if not os.path.isdir('./HITRAN'):
            os.mkdir('./HITRAN')
            print('Downloading molecular data (first time running iris)')
            os.system('python ./Create_ir_model_files.py')
        
        path_to_hitran='./HITRAN'
        self.molecules = molecules   
        self.catalog, self.levels = setup_catalog(molecules, wlow, whigh, path=path_to_hitran)
        
    def setup_disk(self, distance, T_ex, N_mol, A_au, dV):
        """
        setup_disk: setup the disk physical structure

        :distance: distance to source (float)
        :T_ex: excitation temperature (array)
        :N_mol: column density in cm^-2 (array)
        :A_au: emitting area in au^2 (array)
        :dV: intrinsic line FWHM in km/s (array)
        
        The structure of each physical parameter must be an array of 
        shape (M, N), where N is the number of slabs and M the number of species. 
        For now iris does not support using a different number of slabs 
        for each species.
        
        """
        self.distance =  distance
        self.dV = jnp.array(dV)
        self.T_ex = jnp.array(T_ex)
        self.N_mol = jnp.array(N_mol)
        self.A_au = jnp.array(A_au)
        
    def setup_grid(self, fine_wgrid, obs_wgrid, R):
        """
        setup_grid: setup the wavelength grid

        :fine_wgrid: fine wavelength grid to evaluate the opacity
        :obs_wgrid: wavelength grid used to downsample the model
        :R: the instrumental resolving power
        
        ----------------------------------------------------------------
        To appropiately sample overlapping lines, make sure that the 
        wavelength spacing in fine_wgrid appropiately samples the intrinsic
        line width. Ideally you want a few points per 
        intrinsic width, dlambda<1e-5 micron usually works well for MIRI.
        
        To avoid a nan catastrophe, make sure that obs_wgrid does not span
        outside the range of fine_wgrid. 
        """
        
        self.fine_wgrid = jnp.array(fine_wgrid)
        self.obs_wgrid = jnp.array(obs_wgrid)
        # get bin edges 
        grad = jnp.gradient(self.obs_wgrid)
        lis = list(self.obs_wgrid-grad/2)
        lis.append(self.obs_wgrid[-1] + grad[-1]/2)
        self.hist_wgrid = jnp.array(lis)
        # generate histogram weigths
        self.scale_hist, _ = jnp.histogram(self.fine_wgrid, bins=self.hist_wgrid)
        # resolving power
        self.R = R
        # get fine wavelength spacing
        fine_dw = jnp.max(jnp.gradient(self.fine_wgrid))
        mean_w =  ((self.fine_wgrid[-1]+self.fine_wgrid[0])/2)
        # set up convolution window
        dw_R = mean_w / self.R
        sigma_conv = dw_R/fine_dw/2.355
        self.conv_wind = jax.scipy.stats.norm.pdf(jnp.arange(2000)-1000, scale=sigma_conv)
                
    def convolve(self):
        """
        convolution (internal function called by iris.simulate)
        user shouldn't call this directly
        """
        self.convolved_flux = fftconvolve(self.flux_model, self.conv_wind, mode='same')

    def downsample(self):
        """
        downsampling (internal function called by iris.simulate)
        user shouldn't call this directly
        """
        self.downsampled_flux, _ = jnp.histogram(self.fine_wgrid, bins=self.hist_wgrid, weights=self.convolved_flux)
        self.downsampled_flux = self.downsampled_flux / self.scale_hist
        
    def simulate(self):
        """
        Simulate a convolved and downsampled spectrum
        """
        # generate flux density
        self.flux_model = sp.compute_total_fdens(self.catalog, self.distance, self.A_au, 
                                                 self.T_ex, self.N_mol, self.dV, self.fine_wgrid)
        # convolve with instrument psf
        self.convolve()
        # downsample to grid
        self.downsample()




# -*- coding: utf-8 -*-
# HITRAN MOLECULAR DATA PARSER


# --- THIS SECTION ADAPTED FROM CODE WRITTEN BY SIMON BRUDERER (2020) ----
"""
The class MolData reads and parses the molecular data

* Molecular data in the same format as for the Fortran 90 code are used. They consist of a block for the
  partition function and a block for the lines
* Both partition function and lines are stored as named tuples
* Only read access to the fields is granted through properties

- 01/06/2020: SB, initial version

"""

from collections import namedtuple
import numpy as np

try:
    import pandas as pd
except ImportError:
    pd = None
    pass

__all__ = ["MolData"]


class MolData:
    _partition_type = namedtuple('partition', ['t', 'q'])
    _lines_type = namedtuple('lines', ['nr', 'lev_up', 'lev_low', 'lam', 'freq', 'a_stein',
                                       'e_up', 'e_low', 'g_up', 'g_low'])

    def __init__(self, name, fname):
        """Initialize a molecular data structure and read from a .par file. The file has the same structure as in the
        Fortran 90 code, see notes below.

        Parameters
        ----------
        name: str
            Label of the molecule (e.g. "CO")
        fname: str
            Path/filename of the .par file

        Notes
        -----

        The file format of .par file is as follows:

        # Some comments, followed by the number of partition function values
        2933
        # Some further comments, followed by the partition function (temperature vs Q(temperature))
        70.0    25.544319
        71.0    25.904587
        ...
        # Some further comments, followed by the number of lines
        754
        # Some further comments followed by the lines
        1      1_R_13         1_R_13  196.29472  1527.25685208   1.3500e-04     3565.78714     3492.49079   58.0   54.0
        2      0_R_13         0_R_13  194.54560  1540.98815610   2.3660e-04      554.97254      481.01720   58.0   54.0

        Important: The lines section on the bottom is *fixed format*:
        - Nr: Index of the line - integer, 6 characters
        - Lev_up: Upper level label - string, 15 characters
        - Lev_low: Lower level label - string, 15 characters
        - Lam: Wavelength in micron - float, 11 characters
        - Freq: Frequency in GHz - float, 15 characters
        - A_stein: Einstein-A in s**-1 - float, 13 characters
        - E_up: Upper level energy in K - float, 15 characters
        - E_low: Lower level energy in K - float, 15 characters
        - g_up: Upper level statistical weight - float, 7 characters
        - g_low: Lower level statistical weight - float, 7 characters
        """

        self._name = name
        self._fname = fname

        self._partition, self._lines = self._read_molecule(self._fname)

    def _read_molecule(self, fname):
        """Reads the molecular data file, should be only called at initialization

        Parameters
        ----------
        fname: str
            Path/filename of the .par file

        Returns
        -------
        _partition_type:
            Partition function
        _lines_type:
            Lines
        """

        # 1. read file and split into block for partition function and lines
        with open(fname, "r") as f:
            data_raw = f.readlines()

        # remove comments and empty lines
        data_clean = list(filter(lambda x: len(x.strip()) > 0 and x.strip()[0] != "#", data_raw))

        # split file into partition function and line section
        n_partition = int(data_clean[0])

        data_q = data_clean[1:n_partition + 1]
        data_lines = data_clean[n_partition + 2:]

        # 2. read partition function
        N = np.genfromtxt(data_q, dtype="f8,f8")
        q_temperature, q = [N[field] for field in N.dtype.names]

        # 3. read lines

        # noinspection PyTypeChecker
        M = np.genfromtxt(data_lines, dtype="i4,S30,S30,f8,f8,f8,f8,f8,f8,f8",
                          delimiter=(6, 30, 30, 11, 15, 13, 15, 15, 7, 7))

        nr, lev_up, lev_low, lam, freq, a_stein, e_up, e_low, g_up, g_low = [M[field] for field in M.dtype.names]

        lev_up = list(map(self._decode_strip, lev_up))
        lev_low = list(map(self._decode_strip, lev_low))

        # convert frequency from GHz to Hz
        freq = 1e9 * freq

        # 4. return named tuples
        return MolData._partition_type(q_temperature, q), \
            MolData._lines_type(nr, lev_up, lev_low, lam, freq, a_stein, e_up, e_low, g_up, g_low)

    def __repr__(self):
        return f"MolData(name={self.name}, fname={self.fname}, {self.partition}, {self.lines})"

    @staticmethod
    def _decode_strip(x):
        return x.decode().strip()

    @property
    def name(self):
        """str: Name of the molecule"""
        return self._name

    @property
    def fname(self):
        """str: Path/filename of the molecule read in"""
        return self._fname

    @property
    def partition(self):
        """_partition_type: Partition function of the molecule"""
        return self._partition

    @property
    def lines(self):
        """_lines_type: Lines of the molecule"""
        return self._lines

    @property
    def get_table_lines(self):
        """pd.Dataframe: Pandas dataframe with lines table"""

        if pd is None:
            raise ImportError("Pandas required to create table")

        return pd.DataFrame({'nr': self.lines.nr,
                             'lev_up': self.lines.lev_up,
                             'lev_low': self.lines.lev_low,
                             'lam': self.lines.lam,
                             'freq': self.lines.freq,
                             'a_stein': self.lines.a_stein,
                             'e_up': self.lines.e_up,
                             'e_low': self.lines.e_low,
                             'g_up': self.lines.g_up,
                             'g_low': self.lines.g_low})

    @property
    def get_table_partition(self):
        """pd.Dataframe: Pandas dataframe with partition function table"""

        if pd is None:
            raise ImportError("Pandas required to create table")

        return pd.DataFrame({'t': self.partition.t,
                             'q': self.partition.q})

    def _repr_html_(self):
        # noinspection PyProtectedMember
        return self.get_table_lines._repr_html_()
# --- end ----

# Set up with iris (CEM-R)
def setup_catalog(molecule_names, wlow, whigh, wpad=0.01, path='./'):
    path_to_moldata = path+'/'
    # initialize dictionary
    parameters = {}
    levels = {}
    
    for mol_name in molecule_names:
        
        keys = [x.partition('_')[0] for x in parameters.keys()]
        component_count = keys.count(mol_name)
        # initialize dictionary
        mol_name = mol_name + '_{}'.format(component_count)
        parameters[mol_name] = {}
        levels[mol_name] = {}
        
        # read file   
        mol_data = MolData(mol_name.partition('_')[0], path_to_moldata+"data_Hitran_2020_{}.par".format(mol_name.partition('_')[0]))
        # wavelength range
        parameters[mol_name]['wlow'] = wlow-wpad
        parameters[mol_name]['whigh'] = whigh+wpad
        # strucuture the line information
        parameters[mol_name]['ws'] = np.array([w for w in mol_data.get_table_lines['lam']])[::-1]
        levels[mol_name]['lev_up'] = np.array([w for w in mol_data.get_table_lines['lev_up']])[::-1]
        levels[mol_name]['lev_low'] = np.array([w for w in mol_data.get_table_lines['lev_low']])[::-1]
        parameters[mol_name]['aijs'] =  np.array([w for w in mol_data.get_table_lines['a_stein']])[::-1]
        parameters[mol_name]['eups'] = np.array([w for w in mol_data.get_table_lines['e_up']])[::-1] 
        parameters[mol_name]['elow'] = np.array([w for w in mol_data.get_table_lines['e_low']])[::-1] 
        parameters[mol_name]['gups'] = np.array([w for w in mol_data.get_table_lines['g_up']])[::-1] 
        parameters[mol_name]['glow'] = np.array([w for w in mol_data.get_table_lines['g_low']])[::-1] 
        parameters[mol_name]['nus'] = np.array([w for w in mol_data.get_table_lines['freq']])[::-1] 
        # apply wavelength range
        parameters[mol_name]['aijs'] = parameters[mol_name]['aijs'][(parameters[mol_name]['ws']>parameters[mol_name]['wlow'])*(parameters[mol_name]['ws']<parameters[mol_name]['whigh'])]
        levels[mol_name]['lev_up'] = levels[mol_name]['lev_up'][(parameters[mol_name]['ws']>parameters[mol_name]['wlow'])*(parameters[mol_name]['ws']<parameters[mol_name]['whigh'])]
        levels[mol_name]['lev_low'] = levels[mol_name]['lev_low'][(parameters[mol_name]['ws']>parameters[mol_name]['wlow'])*(parameters[mol_name]['ws']<parameters[mol_name]['whigh'])]
        parameters[mol_name]['eups'] = parameters[mol_name]['eups'][(parameters[mol_name]['ws']>parameters[mol_name]['wlow'])*(parameters[mol_name]['ws']<parameters[mol_name]['whigh'])]
        parameters[mol_name]['gups'] = parameters[mol_name]['gups'][(parameters[mol_name]['ws']>parameters[mol_name]['wlow'])*(parameters[mol_name]['ws']<parameters[mol_name]['whigh'])]
        parameters[mol_name]['elow'] = parameters[mol_name]['elow'][(parameters[mol_name]['ws']>parameters[mol_name]['wlow'])*(parameters[mol_name]['ws']<parameters[mol_name]['whigh'])]
        parameters[mol_name]['glow'] = parameters[mol_name]['glow'][(parameters[mol_name]['ws']>parameters[mol_name]['wlow'])*(parameters[mol_name]['ws']<parameters[mol_name]['whigh'])]
        parameters[mol_name]['nus'] = parameters[mol_name]['nus'][(parameters[mol_name]['ws']>parameters[mol_name]['wlow'])*(parameters[mol_name]['ws']<parameters[mol_name]['whigh'])]
        parameters[mol_name]['ws'] = parameters[mol_name]['ws'][(parameters[mol_name]['ws']>parameters[mol_name]['wlow'])*(parameters[mol_name]['ws']<parameters[mol_name]['whigh'])]
        # partition function
        parameters[mol_name]['Qt'] = mol_data.partition[0]
        parameters[mol_name]['Qv'] = mol_data.partition[1]    
        # de-allocate memory
        mol_data = None
    # de-allocate memory    
    mol_data = None
    
    return parameters, levels